/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sGFX_problem.hpp"
#include "sGFX_cube.hpp"
 
//**************************************************************************

SomaProbGFX::SomaProbGFX(SomaProb *const p):SomaSingleGFX(p){
  rotation_allowed = false;
  initCubes();
}

//**************************************************************************

SomaProbGFX::SomaProbGFX(SomaProb *const p,const SomaSingleGFX *const gfx):
                         SomaSingleGFX(p,gfx){
  rotation_allowed = false;
  initCubes();
}

//**************************************************************************

void SomaProbGFX::initCubes(){
  cubes.clear();
  
  if (pbl->get_num_blocks() == 0) {    
    // insert first cube into problem if neccessary
    if (pbl->insert_block(SomaVector<int>())){
      cubes.insert(SomaVector<double>(),SomaVector<double>(0.5,0.5,0.5));
    }
  }else{
    // get all cubes from problem put them into cube array
    offset = pbl->center();

    SomaVector<int> e;
    for(unsigned int i=0;i<pbl->get_num_blocks();i++){
      e = pbl->get_block(i);
      cubes.insert(SomaVector<double>(e)+offset,SomaVector<double>(0.5,0.5,0.5)); 
    }
  } 
  
}

//**************************************************************************

SomaProbGFX::~SomaProbGFX(){}

//**************************************************************************

void SomaProbGFX::display(const GLenum mode){
 
  if (mode == GL_SELECT){
    for(unsigned int i=0;i < cubes.size();i++){
      glPushName(i);
      cubes.get(i)->draw(mode);  
      glPopName();
    }
  }else if (mode == GL_RENDER){
    for(unsigned int i=0;i < cubes.size();i++){
      cubes.get(i)->draw(mode);  
    }    
  }
  
}

//**************************************************************************

void SomaProbGFX::mouseClick(const unsigned int x,const unsigned int y,
                             const MOUSEBUTTON button){
  
  const bool equal = (pbl->get_max_blocks() == pbl->get_num_blocks());
  if( !picking(x,y,button) || equal ) rotation_allowed = true;
  else rotation_allowed = false;
}

//**************************************************************************

void SomaProbGFX::mouseMove(const int deltax,const int deltay,
                            const MOUSEBUTTON button){
   if(button == MIDDLE){ // zoom
    dist += 0.2 * deltay; 
    if(dist < zoomstart-zoomrange) dist = zoomstart-zoomrange;
    if(dist > zoomstart+zoomrange) dist = zoomstart+zoomrange;   
  }else if(button == LEFT && rotation_allowed){ // rotate
    rotation.rotate(SomaMatrix::Y,0.015*deltax);
    rotation.rotate(SomaMatrix::X,0.015*deltay);
  }
}

//**************************************************************************
 
bool SomaProbGFX::picking(const unsigned int x,const unsigned int y,
                          const SomaBaseGFX::MOUSEBUTTON button){
                              
  const unsigned int bufsize = 30*cubes.size();
  unsigned int *select = new unsigned int[bufsize];
  int viewport[4] = {0,0,width,height};
  int hits = 0;
  
  glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);
  
  glSelectBuffer(bufsize,select);
  glRenderMode(GL_SELECT);
  
  glInitNames();   
  
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  
  // create pick matrix, see gluPickMatrix()
  const double sx = viewport[2];
  const double sy = viewport[3];
  const double tx = viewport[2] + 2.0 * -double(x);
  const double ty = viewport[3] + 2.0 * (double(y) - viewport[3]);  
  const double m[16] = {sx,0,0,0,0,sy,0,0,0,0,1,0,tx,ty,0,1};
  
  glLoadMatrixd(m);
  // w,h calculated in resize()
  glFrustum(-w, w, -h, h, 0.5, 0.5+2*(zoomrange+safedist));
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glTranslated(0,0,dist);
  glMultMatrixd(rotation.mtr);
  display(GL_SELECT);
  
  glFlush();
  
  hits=glRenderMode(GL_RENDER);

  if(hits == -1){
#ifdef SOMA_DEBUG    
    std::cerr << "Buffer overflow in SomaProbGFX::picking()." << std::endl;
#endif  
  }else if (hits > 0){
    // parse select array
    unsigned int cpos=0;
    int ip = 0;  
    
    unsigned int numnames = select[cpos];
    unsigned int lowz = select[cpos+1];
    unsigned int name1 = select[cpos+3];
    unsigned int name2 = 0;
    if(numnames > 1) name2 = select[cpos+4];  
    
    cpos += (3+numnames);
    ip++;
    
    while(ip < hits){
      numnames = select[cpos];
      
      if(lowz > select[cpos+1]){
        lowz = select[cpos+1];
        name1 = select[cpos+3];
        if(numnames > 1) name2 = select[cpos+4];
        else name2 = 0;
      }
      
      cpos += (3+numnames);
      ip++;
    }
    
    picked(name1,name2,button);      
  }


  glMatrixMode(GL_PROJECTION); 
  glPopMatrix();

  delete[] select;
 
  return (hits > 0);
}

//**************************************************************************

void SomaProbGFX::picked(const unsigned int name1, const unsigned int name2,
                         const SomaBaseGFX::MOUSEBUTTON button){
  if (button == SomaBaseGFX::LEFT){
    // insert cube at picked side of selected cube
    if (pbl->get_num_blocks() == pbl->get_max_blocks()) return;
      
    SomaVector<double> np = cubes.get(name1)->getPosition();
    SomaVector<int> int_np(np-offset);
  
    switch(name2){
      case SomaCube::XP:
        int_np.x += 1;
        break;
      case SomaCube::XM:
        int_np.x -= 1;
        break;
      case SomaCube::YP:
        int_np.y += 1;
        break;
      case SomaCube::YM:
        int_np.y -= 1;
        break;
      case SomaCube::ZP:
        int_np.z += 1;
        break;
      case SomaCube::ZM:
        int_np.z -= 1;
        break;
    }
    
    if(pbl->insert_block(int_np)) initCubes();
   
  }else if(button == SomaBaseGFX::RIGHT){
    
    // delete picked cube
    if (cubes.size() > 1){
      const SomaVector<double> pos = cubes.get(name1)->getPosition();
      const SomaVector<int> int_pos(pos - offset);
      if (pbl->remove_block(int_pos)) initCubes();
    }
  }
  
}

//**************************************************************************

