/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_piece
#define soma_piece

#include <vector>

#include "s_io.hpp"
#include "s_vector.hpp"

//**************************************************************************

/** around which axis and in which direction to rotate */
enum DIRECTION {ZM=0,ZP,XP,XM,YM,YP,NO};

/** number of orientations we can set a collection to */
const unsigned int ORIENTATIONS=24;
 
//**************************************************************************

/** A class for holding cubes, with 3D-90 rotation functionality. */
template <class T>
class SomaCubeSet{
protected:
  std::vector<SomaVector<T> >* blocks;
  unsigned int curr_or;

public:

  /** create a new collection with one cube at (0,0,0) */
  SomaCubeSet();
  
  /** create a new collection with cubes at (0,0,0) ... (nobin-1,0,0) */
  SomaCubeSet(const unsigned int nobin);
  
  /** create a new collection as copy of another collection */
  SomaCubeSet(const SomaCubeSet& other);
  
  /** delete this collection */
  virtual ~SomaCubeSet();
  
  /** get the number of cubes in this collection */
  inline unsigned int getNumCubes()const{return blocks->size();}
  
  /** set this collection's values to those from another collection */
  void set(const SomaCubeSet& xyz);
  
  /** change coordinates for a given cube */
  void setCube(const unsigned int block,const SomaVector<T>& xyz);

  /** get coordiantes for a given cube */
  SomaVector<T> getCube(const unsigned int block)const;

  /** set the orientation for this collection */
  void setOrientation(const unsigned int orientation);
   
  /** get the orientation of this collection */
  inline unsigned int getOrientation()const{return curr_or;}
  
  /** rotate the vectors into a given direction */
  void rotate(const DIRECTION direction);
  
};

//**************************************************************************

/** a class representing our pieces */
class SomaPiece : public SomaCubeSet<int> , public SomaIO{
private:
  /** check if there is a cube at position (0,0,0) */
  bool static zeroCube(const std::vector<SomaVector<int> > *const cubes);
  /** check for double cubes */
  bool static noDoubleCubes(const std::vector<SomaVector<int> > *const cubes);
  /** check if all cubes are connected */
  bool static connected(const std::vector<SomaVector<int> > *const cubes);

public:
  /** create a new piece with one cube at (0,0,0) */
  inline SomaPiece():SomaCubeSet<int>(){};
  
  /** create a new piece with cubes at (0,0,0) ... (nobin-1,0,0) */
  inline SomaPiece(const unsigned int nobin):SomaCubeSet<int>(nobin){};
  
  /** create a new piece as copy of another piece */
  inline SomaPiece(const SomaPiece& other):SomaCubeSet<int>(other){};
  
  /** delete this piece */
  inline ~SomaPiece(){};
  
  /** get offset to center piece exactly */
  SomaVector<double> getOffset()const;
  
  /** check if this is a valid piece */
  bool valid()const;
  
  /** Write this object's values to a stream. */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read object values from a stream. Returns true if
      the object could be read successfully. */
  bool read(std::istream& strm = std::cin); 
};

//**************************************************************************

#endif

