/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "s_problem.hpp"

#include <iterator>

//**************************************************************************

bool SomaProb::check_solution(const std::vector<SomaPiece> *const pieces,
                              const std::vector<SomaVector<int> > *const cubes,
                              const SomaSolution *const solution){
  
  SomaVector<int> abc,def;
  unsigned int found;
  SomaPiece pc(1);
  std::vector<SomaVector<int> >::const_iterator it;
  
  //compare number of pieces
  if(solution->get_num_pieces() != pieces->size()) return false;
  
  // compare number of cubes
  unsigned int sum = 0;
  for(unsigned int i=0;i<pieces->size();i++){
    if(solution->get(i)){
      sum += (*pieces)[i].getNumCubes();
    }
  }
  if(sum != cubes->size()) return false;
    
  // trying to find all blocks in solution
  it = cubes->begin();
  while(it!= cubes->end()){
    
    // look to which piece this block belongs
    found = 0;
    for(unsigned int i=0;i<pieces->size();i++){
      if(solution->get(i)){
        pc.set((*pieces)[i]);
        pc.setOrientation(solution->get(i)->ori);
        abc = *it - solution->get(i)->pos;
        
        for(unsigned int j=0;j<pc.getNumCubes();j++){
          if(pc.getCube(j) == abc) found++;
        }
        
      }
    }
    
    //whoops .. not found exactly once
    if (found !=1 ) return false;
    
    it++;
  }
  
  //try to find all blocks from solution in block list
  for(unsigned int i=0;i<pieces->size();i++){
    if(solution->get(i)){
      pc.set((*pieces)[i]);
      pc.setOrientation(solution->get(i)->ori);
      
      for(unsigned int j=0;j<pc.getNumCubes();j++){
        def = pc.getCube(j)+solution->get(i)->pos;
        
        found = 0;
        it = cubes->begin();
        while(it != cubes->end()){
          if(*it == def) found++;
          it++;
        }
        
        //not found exactly once
        if(found != 1) return false;
        
      }
      
    }
  }
  
  
  return true;
}

//**************************************************************************

SomaProb::SomaProb(){
  pieces = 0;
  solution = 0;
  cubes = 0;
}

//**************************************************************************

SomaProb* SomaProb::create_problem(std::istream& strm){
  SomaProb* prb = new SomaProb();
  
  if(prb->read(strm)) return prb;
  else{
    delete prb;
    return 0;
  }
  
}

//**************************************************************************

SomaProb::~SomaProb(){
  if(pieces) delete pieces;
  if(cubes) delete cubes;
  if(solution) delete solution;
}

//**************************************************************************

const SomaPiece* SomaProb::get_piece(const unsigned int piece_num)const{
#ifdef SOMA_DEBUG
  if(piece_num >= pieces->size()){
    std::cerr << "Wrong piece number in SomaProb::get_piece()." << std::endl;
    return &(*pieces)[0];
  }
#endif

  return &(*pieces)[piece_num];  
}
  
//**************************************************************************

void SomaProb::add_solution(SomaSolution& solution_in){
#ifdef SOMA_DEBUG
  if(!check_solution(pieces,cubes,&solution_in)){
    std::cerr << "Solution does not solve problem in SomaProb::add_solution()."
              << std::endl;
    return;
  }
#endif
  if(!solution) solution = new SomaSolutions();
  solution->addSolution(solution_in);
}
  
//**************************************************************************

const SomaSolutions* SomaProb::get_solution()const{
  return solution;
}

//**************************************************************************

SomaVector<int> SomaProb::get_block(const unsigned int cube)const{
#ifdef SOMA_DEBUG
  if(cube >= cubes->size()){
    std::cerr << "Cube out of range in SomaProb::get_block()." << std::endl;
    return SomaVector<int>();
  }
#endif
  
  return (*cubes)[cube];
}

//**************************************************************************

bool SomaProb::insert_block(const SomaVector<int>& xyz){
  if (solution) return false;
  
  if((cubes->size() < cubes->capacity())&&(!is_block(xyz))){
    cubes->push_back(xyz);
    return true;
  }else{
    return false;
  }
}

//**************************************************************************

bool SomaProb::remove_block(const SomaVector<int>& xyz){
  if (solution) return false;
  
  std::vector<SomaVector<int> >::iterator it = cubes->begin();
  while(it != cubes->end()){
    if(*it == xyz){
      cubes->erase(it);
      return true;
    }
    it++;
  }
  
  return false;
}

//**************************************************************************

void SomaProb::clear_blocks(){
  if (solution) return;
  cubes->clear();
}

//**************************************************************************

bool SomaProb::is_block(const SomaVector<int>& xyz)const{
  
  std::vector<SomaVector<int> >::const_iterator it = cubes->begin();
  while(it != cubes->end()){
    if(*it == xyz) return true;
    it++;
  }
  
  return false;
}

//**************************************************************************

SomaVector<double> SomaProb::center(){
  std::vector<SomaVector<int> >::iterator it = cubes->begin();
  SomaVector<int> max,min;
  if(it != cubes->end()){
    min = max = *it;
    it++;
  }
  while(it != cubes->end()){
    if(min.x > (*it).x) min.x = (*it).x;
    if(min.y > (*it).y) min.y = (*it).y;
    if(min.z > (*it).z) min.z = (*it).z;

    if(max.x < (*it).x) max.x = (*it).x;
    if(max.y < (*it).y) max.y = (*it).y;
    if(max.z < (*it).z) max.z = (*it).z;
    
    it++;
  }
  
  min += max;
  SomaVector<double> retval(min);
  min /= 2;
  retval /= 2;
  retval -= SomaVector<double>(min);
  
  min.negate();
  retval.negate();
  
  it = cubes->begin();
  while(it != cubes->end()){
    *it += min;
    it++;
  }
  
  if(solution){
    for(unsigned int i=0;i<solution->size();i++){
      solution->get(i)->shift(min);
    }
  }
  
  return retval;
}

//**************************************************************************

void SomaProb::unsolve(){
  if(solution) delete solution;
  solution = 0;
}

//**************************************************************************

void SomaProb::write(std::ostream& strm)const{
  strm << " { (";
  strm.width(3);
  strm << pieces->size() << ") " << std::endl;
  for(unsigned int i=0;i<pieces->size();i++){
    strm << (*pieces)[i];
  }
  
  if(cubes->size() > 0){
    strm << " [ (";
    strm.width(3);
    strm << cubes->size() << ") ";
    std::vector<SomaVector<int> >::const_iterator it = cubes->begin();
    unsigned int count = 0;
    while(it!=cubes->end()){
      if(count%4 == 0) strm << std::endl;
      strm << *it;
      count++;
      it++;
    }
    strm << " ] "; 
  }
  
  if(solution){
    strm << std::endl << " <";
    strm << *solution;
    strm << " > " << std::endl;  
  }
  
  strm << " } " << std::endl;
}

//**************************************************************************

bool SomaProb::read(std::istream& strm){
  char S;
  bool error = false;
  
  unsigned int nw_num_pieces=0;
  std::vector<SomaPiece>* newpieces=0;
  
  SomaSolutions *newsolution=0;

  unsigned int nw_max_blocks=0;
  unsigned int nw_num_blocks=0;
  std::vector<SomaVector<int> >* newblocks = new std::vector<SomaVector<int> >(0);
  
  // read pieces 
  S=' '; 
  strm >> S;
  
  if(S!='{'){ 
    std::cerr << "Error in SomaProb::read(). '{' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S=='('){
    strm >> nw_num_pieces;
    
    if(nw_num_pieces > 0){
      newpieces = new std::vector<SomaPiece>(nw_num_pieces);
    }else{
      std::cerr << "Number of pieces 0 in SomaProb::read()." << std::endl;
      error = true;
      goto error_stop;
    }
  }else{
    std::cerr << "Error in SomaProb::read(). '(' expected" << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S!=')'){
    std::cerr << "Error in SomaProb::read(). ')' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  for(unsigned int i=0;i<newpieces->size();i++){
    if(!(*newpieces)[i].read(strm)){
      error = 1;
      goto error_stop;
    }
  }
  
  for(unsigned int i=0;i<newpieces->size();i++) 
    nw_max_blocks += (*newpieces)[i].getNumCubes();
  
  newblocks->reserve(nw_max_blocks);
    
  S=' '; 
  strm >> S;
    
  if(S=='}'){
    goto error_stop;
  }else if(S!='['){
    std::cerr << "Error in SomaProb::read(). '}' or '[' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }
  
  // read problem 
  S=' '; 
  strm >> S;
  
  if(S=='('){
    strm >> nw_num_blocks;
    
    if(nw_num_blocks == 0 || nw_num_blocks > nw_max_blocks){
      std::cerr << "Incorrect number of blocks in SomaProb::read()."
                << std::endl;
      error = true; 
      goto error_stop;
    }
    
    newblocks->resize(nw_num_blocks);
  }else{
    std::cerr << "Error in SomaProb::read(). '(' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S!=')'){
    std::cerr << "Error in SomaProb::read(). ')' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  for(unsigned int i=0;i<newblocks->size();i++){
    if(!(*newblocks)[i].read(strm)){
      error = true;
      goto error_stop;
    }
  }
  
  S=' '; 
  strm >> S;
  
  if(S!=']'){
    std::cerr << "Error in SomaProb::read(). ']' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  // check for double cubes
  for(unsigned int i=0;i<newblocks->size();i++){
    for(unsigned int j=i+1;j<newblocks->size();j++){
      if((*newblocks)[i] == (*newblocks)[j]){
        std::cerr << "Input sematically incorrect in SomaProb::read()." 
                  << std::endl << "No double blocks allowed." << std::endl;
        error = true; 
        goto error_stop;
      }
    }
  }
  
  S=' '; 
  strm >> S;
      
  if(S=='}'){
    goto error_stop;
  }else if(S!='<'){
      std::cerr << "Error in SomaProb::read(). '}' or '<' expected." 
                << std::endl;
      error = true;
      goto error_stop;
  }
  
  newsolution = new SomaSolutions();
  if( !newsolution->read(strm) ){
    std::cerr << "Could not read solutions in SomaProb::read()." << std::endl;
    error = true;
    goto error_stop;
  }
  
  // check that read solution is indeed a solution
  for(unsigned int i=0;i<newsolution->size();i++){
    if(newsolution->get(i)->get_num_pieces() != newpieces->size()){
      std::cerr << "Error: solution " << i+1 << " has wrong number" 
                << " of pieces in SomaProb::read()." << std::endl;
      error = true;
      goto error_stop;     
    }
    if(!check_solution(newpieces,newblocks,newsolution->get(i))){
      std::cerr << "Error: given solution " << i+1 << "does not " 
                << "solve the problem in SomaProb::read()." << std::endl;
      error = true;
      goto error_stop; 
    }
  }
  
  
  S=' '; 
  strm >> S;
  
  if(S!='>'){
    std::cerr << "Error in SomaProb::read(). '>' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  S=' '; 
  strm >> S;
  
  if(S!='}'){
      std::cerr << "Error in SomaProb::read(). '}' expected." << std::endl;
      error = true;
      goto error_stop;
  }
  
  error_stop:
  
  if(error){
    
    if(newpieces) delete newpieces;
    if(newblocks) delete newblocks;
    if(newsolution) delete newsolution;
    
    return false;
    
  }else{
    
    if(pieces) delete pieces;
    if(cubes) delete cubes;
    if(solution) delete solution;
    
    solution = newsolution;
    pieces = newpieces;

    cubes = newblocks;

    return true;
  }
}

//**************************************************************************

