/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_problem
#define soma_problem

#include <vector>

#include "s_io.hpp"
#include "s_vector.hpp"
#include "s_piece.hpp"
#include "s_solution.hpp"

//**************************************************************************

/** A class for holding piece sets, a figure and solutions */
class SomaProb : public SomaIO{
private:
  std::vector<SomaPiece>* pieces;
  
  SomaSolutions *solution;
  
  std::vector<SomaVector<int> >* cubes;

  SomaProb();
  static bool check_solution(const std::vector<SomaPiece> *const pieces,
                             const std::vector<SomaVector<int> > *const cubes,
                             const SomaSolution *const solution);
public:
  
  /** Create a new problem object from a stream. The stream has to contain
      valid input of course. If reading fails null will be returned. The object
      returned must be deleted later. */ 
  static SomaProb* create_problem(std::istream& strm);
  
  /** Delete this object. */
  ~SomaProb();
  
  /** Get the number of pieces in this problem. */
  inline unsigned int get_num_pieces()const{return pieces->size();}

  /** Get a pointer to a specified piece. */  
  const SomaPiece* get_piece(const unsigned int piece_num)const;
  
  /** Add a solution. */
  void add_solution(SomaSolution& solution_in);
  
  /** Get a pointer to all solutions. If the problem is not solved null will
      be returned. */
  const SomaSolutions* get_solution()const;
  
  /** Get the number of cubes currently inserted into the figure. */
  inline unsigned int get_num_blocks()const{return cubes->size();}
  
  /** Get the maximum number of cubes for a figure. */
  inline unsigned int get_max_blocks()const{return cubes->capacity();}

  /** Get the coordinates for a cube int the figure. */
  SomaVector<int> get_block(const unsigned int cube)const;

  /** Inserts the new given cube if the limit has not been reached and this
      cube is not already inserted. Returns true if insertion was successfull,
      false otherwise. */
  bool insert_block(const SomaVector<int>& xyz);

  /** Removes the given cube if it is in the figure. Returns true if removal
      was successful, false otherwise. */
  bool remove_block(const SomaVector<int>& xyz);

  /** Removes all cubes. */
  void clear_blocks();
  
  /** Check if this cube is in the figure. */
  bool is_block(const SomaVector<int>& xyz)const;
  
  /** Centers the figure on integer coordinates around (0,0,0). Returns the
      offset that needs to be added to each cubes coordinate to center the
      figure completly.*/
  SomaVector<double> center();

  /** Check if problem is currently solved. */
  inline bool is_solved()const{return (solution != 0);}
  
  /** Reset the object after a successfull solve(), so the figure can be edited
      again. */
  void unsolve();
  
  /** Write object values to a stream. This includes the pieces, the figure and
      also all solutions. */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read object values from a stream. The input has to cointain pieces and
      may contain a figure and solutions. Returns true if reading was
      successful, in that case the object values have changed. Otherwise false
      will be returned and the object values will remain. */
  bool read(std::istream& strm = std::cin);
};

//**************************************************************************

#endif

