/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "s_solution.hpp"
#include "s_piece.hpp"

//**************************************************************************

SomaSolution::SomaSolution(const unsigned int num_pieces):
                           posori(num_pieces){
  for(unsigned int i=0;i<posori.size();i++)
    posori[i].ori = ORIENTATIONS;  
}

//**************************************************************************

SomaSolution::SomaSolution(const SomaSolution& solution_in):
                           posori(solution_in.posori){

}

//**************************************************************************
  
SomaSolution::~SomaSolution(){

}

//**************************************************************************  

void SomaSolution::insert(const unsigned int piece,const SomaPosOri& po){
#ifdef SOMA_DEBUG
  if(piece >= posori.size()){
    std::cerr << "Piece out of range in SomaSolution::insert()" << std::endl;
    return;
  }
  if(po.ori >= ORIENTATIONS){
    std::cerr << "Orientation out of range in SomaSolution::insert()"
              << std::endl;
    return;
  }
#endif

  posori[piece] = po;
}

//**************************************************************************  

void SomaSolution::remove(const unsigned int piece){
#ifdef SOMA_DEBUG
  if(piece >= posori.size()){
    std::cerr << "Piece out of range in SomaSolution::remove()" << std::endl;
    return;
  }
#endif

  posori[piece].ori = ORIENTATIONS;
}

//**************************************************************************

void SomaSolution::clear(){
  for(unsigned int i=0;i<posori.size();i++){
    posori[i].ori = ORIENTATIONS;
  }
}

//**************************************************************************  

void SomaSolution::shift(const SomaVector<int>& offset){
  for(unsigned int i=0;i<posori.size();i++){
    if(posori[i].ori != ORIENTATIONS){
      posori[i].pos += offset;
    }
  }
}

//**************************************************************************  

const SomaPosOri* SomaSolution::get(const unsigned int piece)const{
#ifdef SOMA_DEBUG
  if(piece >= posori.size()){
    std::cerr << "Piece out of range in SomaSolution::get()" << std::endl;
    return false;
  }
#endif
  
  if(posori[piece].ori == ORIENTATIONS) return 0;
  return &posori[piece];
}

//**************************************************************************  

void SomaSolution::write(std::ostream& strm)const{
  strm << " {";
  strm << " (";
  strm.width(3);
  strm << posori.size();
  strm << ")";
  for(unsigned int i=0;i<posori.size();i++){
    strm << " (";
    strm.width(3);
    strm << posori[i].ori;
    strm << ")";
    strm << posori[i].pos;
    if(!((i+1) % 3)) strm << std::endl;
  }
  strm << " }" << std::endl;
}

//**************************************************************************  

bool SomaSolution::read(std::istream& strm){
  char S=' ';
  
  S = ' ';
  strm >> S;
  if(S != '{'){
    std::cerr << "Error in SomaSolution::read(). '{' expected." << std::endl;
    return false;
  } 
  
  
  S = ' ';
  strm >> S;
  if(S != '('){
    std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
    return false;
  }
  
  unsigned int new_num_pieces=0;
  strm >> new_num_pieces;
  if(new_num_pieces == 0){
    std::cerr << "Error in SomaSolution::read(). Number of pieces 0." 
              << std::endl;
    return false;  
  }
  
  S = ' ';
  strm >> S;
  if(S != ')'){
    std::cerr << "Error in SomaSolution::read(). ')' expected." << std::endl;
    return false;
  }
  
  posori.resize(new_num_pieces);
  
  for(unsigned int i=0;i<posori.size();i++){
    S = ' ';
    strm >> S;
    if(S != '('){
      std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
      return false;
    }else{
      strm >> posori[i].ori;
      if(posori[i].ori > ORIENTATIONS){
        std::cerr << "Error in SomaSolution::read(). Orientation "
                  << posori[i].ori << " out of range." << std::endl;
        return false;
      }
    }
    
    S = ' ';
    strm >> S;
    if(S != ')'){
      std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
      return false;  
    }else{
      if(! posori[i].pos.read(strm)) return false;
    }
    
  }
  
  S = ' ';
  strm >> S;
  if(S != '}'){
    std::cerr << "Error in SomaSolution::read(). '}' expected." << std::endl;
    return false;
  } 
  
  return true;
}

//**************************************************************************
//**************************************************************************
//**************************************************************************

SomaSolutions::SomaSolutions(){
  active_solution = 0;
}
  
//**************************************************************************

SomaSolutions::~SomaSolutions(){
  for(unsigned int i=0;i<solutions.size();i++){
    delete solutions[i];
  }
}
  
//**************************************************************************

void SomaSolutions::addSolution(const SomaSolution& solution){
  solutions.push_back(new SomaSolution(solution));
  active_solution = 0;
}

//**************************************************************************

void SomaSolutions::nextSolution()const{
  if(active_solution == solutions.size()-1) active_solution = 0;
  else active_solution++;
}

//**************************************************************************

void SomaSolutions::prevSolution()const{
  if(active_solution == 0) active_solution = solutions.size()-1;
  else active_solution--;
}  

//**************************************************************************

const SomaSolution* SomaSolutions::get(const unsigned int number)const{
#ifdef SOMA_DEBUG
 if(number >= solutions.size()){
   std::cerr << "Number out of range in SomaSolutions::get()" << std::endl;
   return 0;
 }
#endif

 return solutions[number];
}
  
//**************************************************************************

SomaSolution* SomaSolutions::get(const unsigned int number){
#ifdef SOMA_DEBUG
 if(number >= solutions.size()){
   std::cerr << "Number out of range in SomaSolutions::get()" << std::endl;
   return 0;
 }
#endif

 return solutions[number];
}
  
//**************************************************************************

void SomaSolutions::write(std::ostream& strm)const{
  strm << " (";
  strm.width(3);
  strm << solutions.size();
  strm << ")" << std::endl;
  for(unsigned int i=0;i<solutions.size();i++){
    solutions[i]->write(strm);
  }
}
  
//**************************************************************************

bool SomaSolutions::read(std::istream& strm){
  char S=' ';
  
  unsigned int num_solutions = 0;
  std::vector<SomaSolution* > readsolutions;
  
  strm >> S;
  if(S != '('){
    std::cerr << "Error in SomaSolutions::read(). '(' expected." << std::endl;
    return false;
  }
  
  strm >> num_solutions;
  if(num_solutions == 0){
    std::cerr << "Error in SomaSolutions::read(). Number of solutions is 0." 
              << std::endl;
    return false;
  }
  
  S = ' ';
  strm >> S;
  if(S != ')'){
    std::cerr << "Error in SomaSolutions::read(). ')' expected." << std::endl;
    return false;
  }
  
  bool noerror = true;
  for(unsigned int i=0;i<num_solutions;i++){
    SomaSolution *readsolution = new SomaSolution(1);
    noerror = readsolution->read(strm);
    readsolutions.push_back(readsolution);
    if(!noerror) break;
  }
  
  if(!noerror){
    for(unsigned int i=0;i<readsolutions.size();i++){
      delete readsolutions[i];
    }
    return false;
  }else{
    for(unsigned int i=0;i<solutions.size();i++){
      delete solutions[i];
    }
    solutions.clear();
    for(unsigned int i=0;i<readsolutions.size();i++){
      solutions.push_back(readsolutions[i]);
    }
    active_solution = 0;
    return true;
  }

}
  
//**************************************************************************

