/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_solution
#define soma_solution

#include <vector>

#include "s_io.hpp"
#include "s_vector.hpp"


//**************************************************************************

/** Combine position and orientation for an inserted piece */
struct SomaPosOri{
  /** position */
  SomaVector<int> pos;
  /** orientation */
  unsigned int ori;  
};

//**************************************************************************

/** a container holding one solution with read/write capability */
class SomaSolution : public SomaIO{
private:
  std::vector<SomaPosOri> posori;
public:

  /** Constructor, create a solution object for num_pieces pieces. */
  SomaSolution(const unsigned int num_pieces);
  
  /** Copy constructor. */
  SomaSolution(const SomaSolution& solution_in);
  
  /** Destructor. */
  ~SomaSolution();
  
  inline unsigned int get_num_pieces()const{ return posori.size(); }
  
  /** Insert a piece into solution, save position and orientation of this
      piece. */
  void insert(const unsigned int piece,const SomaPosOri& posori);

  /** Remove a piece from solution. */
  void remove(const unsigned int piece);
  
  /** Remove all pieces from solution. */
  void clear();
  
  /** Add's offset to each position. */
  void shift(const SomaVector<int>& offset);
  
  /** Get position and orientation for a piece, returns true if the piece is
      in the solution, false otherwise */
  const SomaPosOri* get(const unsigned int piece)const;

  /** Write solution to a stream. */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read solution from a stream. Returns true if reading was successful, in
      that case the object values have changed. Otherwise false will be
      returned and the object values are undefined. */
  bool read(std::istream& strm = std::cin);
};

//**************************************************************************

/** a container holding solutions with read/write capability */
class SomaSolutions : public SomaIO{
private:
  std::vector<SomaSolution *> solutions;
  mutable unsigned int active_solution;
public:

  /** constructor */
  SomaSolutions();
  
  /** destructor */
  ~SomaSolutions();
  
  /** add one solution to container */
  void addSolution(const SomaSolution& solution);
  
  /** return number of solutions in container */
  inline unsigned int size()const{ return solutions.size(); }
  
  /** set active solution to next solution in instance, this will circulate */
  void nextSolution()const;
  
  /** set active solution to previos solution in instance, this will
      circulate */
  void prevSolution()const;
  
  /** get const pointer to one solution, valid numbers are 0..size()-1 */
  const SomaSolution* get(const unsigned int number)const;
  
  /** get const pointer to currently active solution */
  inline const SomaSolution* get()const{ return get(active_solution); }
  
  /** get pointer to one solution, valid numbers are 0..size()-1 */
  SomaSolution* get(const unsigned int number);

  /** get pointer to currently active solution */
  inline SomaSolution* get(){ return get(active_solution); }
  
  /** return the currently active solution */
  inline unsigned int getActiveSolution()const{ return active_solution; }
 
  /** Write solutions to a stream. */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read solutions from a stream. Returns true if reading was successful, in
      that case the object values have changed. Otherwise false will be
      returned and the object values will remain. */
  bool read(std::istream& strm = std::cin);  
  
};

#endif

