/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdlib>
#include <ctime>

#include <iterator> 
 
#include "s_solve.hpp" 
#include "s_vector.hpp"
#include "s_piece.hpp"

//**************************************************************************

bitvector SomaSolve::flags[8*sizeof(bitvector)+1];

//**************************************************************************

void SomaSolve::initFlags(){
  static unsigned int stop = 0;
  if(stop) return;
  stop = 1;
#ifdef SOMA_VERBOSE
  std::cout << "Bitvector has " << 8*sizeof(bitvector) << " bits." << std::endl;
#endif
  flags[0] = 0;
  flags[1] = 1;
  for(unsigned int i=1;i<8*sizeof(bitvector);i++){
    flags[i+1] = 2*flags[i];
  }
}

//**************************************************************************

void SomaSolve::printT(const bitvector val){
  for(unsigned int i=8*sizeof(bitvector);i>0;i--){
    if(val & flags[i]) std::cout << "#";
    else std::cout << "_";
    if((i-1)%8 == 0 && i!=1) std::cout << " ";
  }
  std::cout << std::endl;
}
  
//**************************************************************************

unsigned int SomaSolve::calc_blockmax(const SomaProb *const p){
  // how big is the largest piece ?
  // this will determine the size of our 3d space to prevent array overflows
  unsigned int blockmax = p->get_piece(0)->getNumCubes();
  for(unsigned int i=0;i<p->get_num_pieces();i++){
    if(p->get_piece(i)->getNumCubes() > blockmax) 
      blockmax =  p->get_piece(i)->getNumCubes();
  }
  // we need at least a "safety border" of one block to prevent array overflows 
  // in the connectCheck, otherwise we may decrease it by one
  if(blockmax > 1) blockmax--;
  
  return blockmax;
}
  
//**************************************************************************

SomaVector<int> SomaSolve::calc_min(const SomaProb *const p){
  SomaVector<int> block,min;
  if(p->get_num_blocks() > 0){
    block = p->get_block(0);
    min.x = block.x;
    min.y = block.y;
    min.z = block.z;
  }
  for(unsigned int i=1;i<p->get_num_blocks();i++){
    block = p->get_block(i);
    
    if(block.x < min.x) min.x = block.x;
    if(block.y < min.y) min.y = block.y;
    if(block.z < min.z) min.z = block.z;
  }

  return min;  
}
  
//**************************************************************************

SomaVector<int> SomaSolve::calc_max(const SomaProb *const p,const SomaVector<int>& min,const unsigned int blockmax){
  SomaVector<int> block,max;
  if(p->get_num_blocks() > 0){
    block = p->get_block(0);
    max.x = block.x;
    max.y = block.y;
    max.z = block.z;
  }
  for(unsigned int i=1;i<p->get_num_blocks();i++){
    block = p->get_block(i);
    
    if(block.x > max.x) max.x = block.x;
    if(block.y > max.y) max.y = block.y;
    if(block.z > max.z) max.z = block.z;
  }
  
  max.x = max.x - min.x + 2*blockmax + 1;
  max.y = max.y - min.y + 2*blockmax + 1;
  max.z = max.z - min.z + 2*blockmax + 1;

  return max;  
}
  
//**************************************************************************

SomaSolve* SomaSolve::createSolver(SomaProb *const p){
 if(p->get_num_blocks() == 0) return 0;
 if(p->get_num_blocks() > 8*sizeof(bitvector)) return 0;
 for(unsigned int i=0;i<p->get_num_pieces();i++){
   if(!p->get_piece(i)->valid()) return 0;
 }
 
 return new SomaSolve(p);
}
  
//**************************************************************************
 
SomaSolve::SomaSolve(SomaProb *const p_in):p(p_in),
                                           num_blocks(p->get_num_blocks()),
                                           num_pieces(p->get_num_pieces()),
                                           blockmax(calc_blockmax(p)),
                                           min(calc_min(p)),
                                           max(calc_max(p,min,blockmax)),
                                           xmult(max.y*max.z),
                                           ymult(max.z),
                                           val2(xmult << 1),
                                           val3(xmult+ymult),
                                           val4(ymult << 1),
                                           val5(ymult+1),
                                           cubes(num_blocks){
 
  initFlags();
  
  // connectCheck Stuff
  queue = new unsigned int[num_blocks];
  confound = new unsigned int[num_blocks+1];
  reverse_lookup = new int[num_blocks+1];
  confound[0] = false; // 0 blocks are always found already
  
  knapsack_solvable = new unsigned int[num_pieces*num_blocks];
  for(unsigned int i=0;i<num_pieces;i++){
    for(unsigned int j=0;j<num_blocks;j++){
      knapsack_solvable[i*num_blocks+j] = 0;
    }
  }
  
  // coverCheck Stuff
  cover = new bitvector[num_pieces];
  
  //create array out of blocklist
  space = new unsigned int[max.x*max.y*max.z];

  for(int i=0;i<max.x;i++){
    for(int j=0;j<max.y;j++){
      for(int k=0;k<max.z;k++){
        space[i*xmult + j*ymult + k] = 0;
      }
    }
  }
  
  
  //number blocks 1 ... num_blocks
  SomaVector<int> block;
  for(unsigned int i=0;i<p->get_num_blocks();i++){
    block = p->get_block(i);
    space[(block.x-min.x+blockmax) * xmult 
        + (block.y-min.y+blockmax) * ymult 
        + (block.z-min.z+blockmax)] = i+1;
        
    reverse_lookup[i+1] = (block.x-min.x+blockmax) * xmult 
                          + (block.y-min.y+blockmax) * ymult 
                          + (block.z-min.z+blockmax);
  }
 
  createLists();
  
  
  // stuff for main iteration  
  inuse = new bitvector[num_pieces];
  
  ial2 = new unsigned int[num_pieces];
  insarr2 = new bitvector*[num_pieces];
  iapn2 = new unsigned int[num_pieces];
  
  solved = 0;
  for(unsigned int i=1;i<=num_blocks;i++){
    solved |= flags[i];
  }
  
  //to determine which pieces are possible solve knapsack problem
  unsigned int* a = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    a[i] = p->get_piece(iapn[i])->getNumCubes();
  }
  K = new SomaKnapsack(a,num_blocks,num_pieces);
  delete[] a;
  
  //for rotation check get copy of cubes
  SomaVector<double> offset = p->center(); // assume figure is centered already
  for(unsigned int i=0;i<num_blocks;i++){
    cubes.setCube(i,SomaVector<double>(p->get_block(i))+offset);
  }
  permutation = new unsigned int[num_blocks+1];
}

//**************************************************************************

SomaSolve::~SomaSolve(){

  delete[] space;
  
  delete[] insmap;
  for(unsigned int i=0;i<num_pieces;i++) delete[] insarr[i];
  delete[] insarr;
  
  delete[] ial;

  delete[] inuse;
  delete K;
  
  // connectCheck Stuff
  delete[] confound;
  delete[] queue;
  delete[] reverse_lookup;
  delete[] knapsack_solvable;
  
  // coverCheck Stuff
  delete[] cover;
  
  delete[] iapn;
  
  delete[] iapn2;
  delete[] ial2;
  delete[] insarr2;

  delete[] permutation;
}

//**************************************************************************
 
void SomaSolve::createLists(){
  // lists containing for each piece each position it could be
  // inserted and which blocks it fills then 
  insmap = new std::map<bitvector,SomaPosOri>[num_pieces];
  
  SomaVector<int> pos;
  SomaPiece piece(1);
  unsigned int temp;
  
  // check for each piece how it fits into the blocks, for each insertion 
  // save which blocks are filled in a list
  for(unsigned int i=0;i<num_pieces;i++){
    piece.set(*p->get_piece(i));
    const unsigned int blocknum = piece.getNumCubes();
    SomaPosOri current;
    bitvector currentvec=0;

    for(unsigned int j=0;j<ORIENTATIONS;j++){
      piece.setOrientation(j);
      
      const SomaVector<int> max2(max.x-blockmax,max.y-blockmax,max.z-blockmax);
      for(int l=blockmax;l<max2.x;l++){
        for(int m=blockmax;m<max2.y;m++){
          for(int n=blockmax;n<max2.z;n++){
            // try to insert piece at l,m,n
            
            for(unsigned int k=0;k<blocknum;k++){
              pos = piece.getCube(k);
              pos.x+=l; pos.y+=m; pos.z+=n;
              
              if(space[pos.x*xmult + pos.y*ymult + pos.z] != 0){
                currentvec |= flags[space[pos.x*xmult + pos.y*ymult + pos.z]];
              }else{
                currentvec = 0;
                break;
              }
            }
            
            if(currentvec){
              current.pos.set(l-blockmax,m-blockmax,n-blockmax);
              current.pos += min;
              current.ori = j;
  
              insmap[i][currentvec]=current;

              currentvec = 0;
            }      
          }
        }
      }
    }
  }
 
  //generate sorted index
  iapn = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    iapn[i]  = i;
  }
  
  // sort fore piecesize
  for(unsigned int i=1;i<num_pieces;i++){
    temp = iapn[i];
    unsigned int j = i;
    while(j>0 && p->get_piece(temp)->getNumCubes() <  p->get_piece(iapn[j-1])->getNumCubes()){
      iapn[j] = iapn[j-1];
      j--;
    }
    iapn[j] = temp;
  }
  
  // sort for listlength
  for(unsigned int i=1;i<num_pieces;i++){
    temp = iapn[i];
    unsigned int j = i;
    while(j>0 && insmap[temp].size() < insmap[iapn[j-1]].size()){
      iapn[j] = iapn[j-1];
      j--;
    }
    iapn[j] = temp;
  }
  
  //make copy of insarr to insarr
  std::map<bitvector,SomaPosOri>::const_iterator it;
  insarr = new bitvector*[num_pieces];
  ial = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    ial[i] = insmap[iapn[i]].size();
    insarr[i] = new bitvector[ial[i]];
    it = insmap[iapn[i]].begin();
    for(unsigned int j=0;j<ial[i];j++){
      insarr[i][j] = it->first;
      it++;
    }
  }

}
                   
//**************************************************************************

void SomaSolve::initConnectCheck(){

#ifdef SOMA_VERBOSE  
  for(unsigned int i=0;i+1<num_selected;i++){
    for(unsigned int b=0;b<num_blocks;b++){
      knapsack_solvable[i*num_blocks+b] = 0;
    }
  }
#endif
  
  unsigned int dim = num_selected;
  
  unsigned int* a = new unsigned int[dim];
  unsigned int j = dim-1;
  
  for(unsigned int i=0;i<num_selected;i++){
    a[j--] = p->get_piece(iapn2[i])->getNumCubes();      
  }
  
  j = num_blocks;
  SomaKnapsack k(a,1,dim);
  for(unsigned int i=0;i<num_selected;i++){
    j -= a[--dim];
    if(dim > 0){
      k.set_dim(dim);
      for(unsigned int b=1;b<=j;b++){
        k.set_b(b);
        knapsack_solvable[(i)*num_blocks+b] = (k.next_solution() ? 1 : 0);
      }
    }
  }

  delete[] a;
#ifdef SOMA_VERBOSE  
  for(unsigned int i=0;i+1<num_selected;i++){
    for(unsigned int b=0;b<num_blocks;b++){
      std::cout << (knapsack_solvable[i*num_blocks+b] ? "#" : "_");
    }
    std::cout << std::endl;
  }
#endif

}

//**************************************************************************

unsigned int SomaSolve::connectCheck(){ 
  unsigned int concurr; 
  unsigned int cindex; 
  unsigned int qs=0,qe=0; 

  unsigned int *pi, i;
  
  // mark inserted blocks as found 
  for(i=1;i<=num_blocks;i++){ 
    confound[i] = !(inserted & flags[i]); 
  } 

  // check open problem space for connected components with bfs   
  // search start point 
  for(i=num_blocks;i>0;i--){ 
     
    if(confound[i]){        
      // new component found  
      queue[qs++] = reverse_lookup[i]; 
      confound[i] = false; 
      concurr = 0; 
       
      // simple bfs 
      while(qs>qe){ 
        concurr++; 
         
        // check in six directions
        
        cindex = queue[qe++] + xmult; 
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex -= val2;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
        cindex += val3;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
        cindex -= val4;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex += val5;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex -= 2;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
      } 

      // open space can be filled in knapsack relaxation ??
      if(! knapsack_solvable[(ta)*(num_blocks)+concurr]) return true;
      
    } 
     
  } 
    
  return false; 
}

//**************************************************************************

unsigned int SomaSolve::coverCheck(){
  bitvector cover1 = inserted;
  for(unsigned int i=ta+1;i<num_selected;i++){
    for(unsigned int j=0;j<ial2[i];j++){
      if(! (insarr2[i][j] & inserted)){
        cover1 |= insarr2[i][j];
      }
    }
  }
  
  cover[ta] = cover1;
  for(unsigned int j=0;j<ial2[ta];j++){
    if(! (insarr2[ta][j] & inserted)){
      cover1 |= insarr2[ta][j];
    }
  }
  return !(cover1 == solved); 
}

//**************************************************************************

void SomaSolve::recurse(){

  if(coverCheck()) return;
  
  for(unsigned int i=0;i<ial2[ta];i++){
    if(!(insarr2[ta][i] & inserted) && ((insarr2[ta][i] | cover[ta]) == solved)){
      inserted |= insarr2[ta][i];
      inuse[ta] = insarr2[ta][i];
      
      if(inserted == solved){

#ifdef SOMA_VERBOSE
        std::cout.width(4);
        std::cout << i+1 << "\b\b\b\b" << std::flush;
#endif  
      
        if(!just_count) writeSolution();
        if(++num_solutions >= stop_solutions) return;
            
      }else if(!connectCheck()){
        ta++;

#ifdef SOMA_VERBOSE
        std::cout.width(4);
        std::cout << i+1 << std::flush;
#endif  
        recurse();
#ifdef SOMA_VERBOSE
        std::cout << "\b\b\b\b";
#endif  

        ta--;
        if(num_solutions >= stop_solutions) return;
      }
      
      inserted ^= insarr2[ta][i];
    }
  } 
}

//**************************************************************************

unsigned int SomaSolve::getPermutation(const SomaCubeSet<double>& source,
                                       const SomaCubeSet<double>& target,
                                       unsigned int *const permutation){
  // we know there are no double cubes, so check if each one from source is
  // also in target
  for(unsigned int i=0;i<num_blocks;i++){
    if(source.getCube(i) == SomaVector<double>()){
      permutation[i+1] = i+1;
    }else{
      permutation[i+1] = 0;
      for(unsigned int j=0;j<num_blocks;j++){
        if(source.getCube(i) == target.getCube(j)){
          permutation[i+1] = j+1;
          break;
        }
      }
      if(!permutation[i+1]) return 0;
    }
  }
  
  return 1;                                       
}

//**************************************************************************

void SomaSolve::shortenList(std::list<bitvector>& inslist,
                            const unsigned int *const permutation){
  bitvector kill;
  std::list<bitvector>::iterator it,jt;
  for(it=inslist.begin();it != inslist.end();it++){
    kill=0;
    for(unsigned int i=1;i<=num_blocks;i++)
      if(*it & flags[i]) kill |= flags[permutation[i]];
    
    jt=it; 
    jt++;
    while(jt != inslist.end()){
      if(*jt == kill) jt = inslist.erase(jt);
      else jt++;
    }
  }
}

//**************************************************************************
                   
void SomaSolve::recurseSymmetryCheck(){

  if(coverCheck()) return;
  
  std::list<bitvector> inslist;
  for(unsigned int i=0;i<ial2[ta];i++) inslist.push_back(insarr2[ta][i]);
  
  SomaCubeSet<double> source(cubes);
  
  for(unsigned int i=1;i<=num_blocks;i++)
    if(inserted & flags[i])
      source.setCube(i-1,SomaVector<double>());
  
  SomaCubeSet<double> target(source);
  
  unsigned int pf = 0;
  for(unsigned int i=1;i<ORIENTATIONS;i++){
    target.setOrientation(i);
    if(getPermutation(source,target,permutation)){
      pf++;
      shortenList(inslist,permutation);
    }
  }
  if(pf && (inslist.size()*(pf+1) == ial2[ta])) pf = 0;
  
  
#ifdef SOMA_VERBOSE
  unsigned int j=0;
#endif

  for(std::list<bitvector>::const_iterator i=inslist.begin();i!=inslist.end();i++){

    if(!(*i & inserted) && ((*i | cover[ta]) == solved)){
      inserted |= *i;
      inuse[ta] = *i;
      
      if(inserted == solved){

#ifdef SOMA_VERBOSE
        std::cout.width(4);
        std::cout << j+1 << "\b\b\b\b" << std::flush;
#endif  
      
        if(!just_count) writeSolution();
        if(++num_solutions >= stop_solutions) return;
            
      }else if(!connectCheck()){
        ta++;

#ifdef SOMA_VERBOSE
        std::cout.width(4);
        std::cout << j+1 << std::flush;
#endif  
        if(pf) recurseSymmetryCheck();
        else recurse();
#ifdef SOMA_VERBOSE
        std::cout << "\b\b\b\b";
#endif  

        ta--;
        if(num_solutions >= stop_solutions) return;
      }
      
      inserted ^= *i;
    }
#ifdef SOMA_VERBOSE
    j++;
#endif
  } 
}

//**************************************************************************

unsigned int SomaSolve::solve(unsigned int stop_solutions_in,
                              const unsigned int randomize,
                              const unsigned int just_count_in){
  
  stop_solutions = stop_solutions_in;
  just_count = just_count_in;
  
  if(randomize){
    srand(time(0));
    for(unsigned int i=0;i<num_pieces;i++){
      for(unsigned j=0;j<ial[i];j++){
        const unsigned int k = rand()%ial[i];
        bitvector swap = insarr[i][j];
        insarr[i][j] = insarr[i][k];
        insarr[i][k] = swap;
      }
    }
  }
  
  num_solutions = 0;

  possible = K->next_solution();
  while(possible){    
    
    num_selected = 0;
    for(unsigned int i=0;i<num_pieces;i++){
      if(possible[i]){
        insarr2[num_selected] = insarr[i];
        ial2[num_selected] = ial[i];
        iapn2[num_selected] = iapn[i];
        num_selected++;
      }
    }
    
    inserted = 0;
    ta = 0;
    initConnectCheck();
    
#ifdef SOMA_VERBOSE
    std::cout << "       ";
    for(unsigned int j=0;j<num_pieces;j++){
      if(possible[j]){
        std::cout.width(4);
        std::cout << j;
      }
    }
    std::cout << std::endl << "size  :";
    for(unsigned int j=0;j<num_selected;j++){
      std::cout.width(4);
      std::cout << p->get_piece(iapn2[j])->getNumCubes();
    }
    std::cout << std::endl << "lenght:";
    for(unsigned int j=0;j<num_selected;j++){
        std::cout.width(4);
        std::cout << ial2[j];
    }
    std::cout << std::endl << "pos   :";     
#endif
    
    recurseSymmetryCheck();

#ifdef SOMA_VERBOSE
    std::cout << std::endl;
#endif
    
    if(num_solutions >= stop_solutions) return num_solutions;
 
    possible = K->next_solution();  
  }
  
  return num_solutions;
}

//**************************************************************************

void SomaSolve::writeSolution(){
  SomaSolution solution(num_pieces);
 
  for(unsigned int j=0,i=0;i<num_pieces;i++){
    if(possible[i]){
      solution.insert(iapn2[j],insmap[iapn2[j]][inuse[j]]);
      j++;
    }
  }
  p->add_solution(solution);
}

//**************************************************************************

