/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_vector
#define soma_vector

#include <cmath>

#include "s_io.hpp"

//**************************************************************************

/** return absolute value */
template <class T>
inline T somaAbs(const T val){
  return 0 < val ? val : -val;
}

/** A simple class handling x,y,z coordinates. */
template <class T> class SomaVector : public SomaIO{
  public:
  T x,y,z; 
  
  /** Create a new Coordinate ( 0 , 0 , 0 ). */
  inline SomaVector(){
    x = y = z = 0;
  }
  
  /** Create a new coordinate with start values x, y, z. */
  inline SomaVector(const T xin, const T yin, const T zin){
    x = xin;
    y = yin;
    z = zin;
  }
  
  /** Create a new coordinate with values from another coordinate. */
  inline SomaVector(const SomaVector<int>& xyz){
    x = (T) xyz.x;
    y = (T) xyz.y;
    z = (T) xyz.z;
  }
  
   /** Create a new coordinate with values from another coordinate. */
  inline SomaVector(const SomaVector<double>& xyz){
    x = (T) xyz.x;
    y = (T) xyz.y;
    z = (T) xyz.z;
  }
  
  /** Delete this object. */
  inline ~SomaVector(){}
  
  /** Set the values to given x, y, z. */
  inline void set(const T xin, const T yin, const T zin){
    x = xin;
    y = yin;
    z = zin;
  }
  
  /** Add given values to the coordinate */
  inline void add(const T xin, const T yin, const T zin){
    x += xin;
    y += yin;
    z += zin;
  }
  
  /** Add values from another coordinate to this coordinate. */
  inline void operator += (const SomaVector<T>& xyz){
    x += xyz.x;
    y += xyz.y;
    z += xyz.z;
  }
  
  /** Subtract given values from the coordinate */
  inline void subtract(const T xin, const T yin, const T zin){
    x -= xin; 
    y -= yin; 
    z -= zin;
  }
  
  /** Subtract values from another coordinate from this coordinate */
  inline void operator -= (const SomaVector<T>& xyz){
    x -= xyz.x;
    y -= xyz.y;
    z -= xyz.z;
  }
  
  /** Scale vector */
  inline void operator *= (const T scl){
    x *= scl;
    y *= scl;
    z *= scl;
  }
  
  /** Scale vector */
  inline void operator /= (const T scl){
    x /= scl;
    y /= scl;
    z /= scl;
  }
  
  /** Compare coordinate for equality with given values. */
  inline bool equal(const T xin, const T yin, const T zin)const{
    if(x != xin) return false;
    if(y != yin) return false; 
    if(z != zin) return false;
    return true;
  }
  
  /** Compare two coordiantes for equality. */
  inline bool operator == (const SomaVector<T>& xyz)const{
    if(x != xyz.x) return false;
    if(y != xyz.y) return false;
    if(z != xyz.z) return false;
    return true;
  }
    
  /** return the length of this vector in the one norm */  
  inline T oneNorm()const{
    return somaAbs(x)+somaAbs(y)+somaAbs(z);
  }
  
  /** return the length of this vector in the two or euclidean norm */
  inline T twoNorm()const{
    return (T) sqrt((double)(x*x+y*y+z*z));
  }
  
  /** return the length of this vector in the infinity norm */
  inline T infNorm()const{
    const T tmp = somaAbs(x) > somaAbs(y) ? somaAbs(x) : somaAbs(y);
    return somaAbs(z) > tmp ? somaAbs(z) : tmp;
  }

  /** negate the vector */
  inline void negate(){
    x = -x;
    y = -y;
    z = -z;
  }
  
  /** Write object values to a stream */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read new values for this object from a stream.
  True will be returned if reading was successfull
  and object values have been changed. Otherwise
  false will be returned and the x, y, z values
  become undefined */
  bool read(std::istream& strm = std::cin);
  
};

//**************************************************************************

/** a simple addition operator */
template <class T>
inline SomaVector<T> operator + (const SomaVector<T>& a,
                                 const SomaVector<T>& b){
  return SomaVector<T>(a.x+b.x,a.y+b.y,a.z+b.z);
}

//**************************************************************************

/** a simple subtraction operator */
template <class T>
inline SomaVector<T> operator - (const SomaVector<T>& a,
                                 const SomaVector<T>& b){
  return SomaVector<T>(a.x-b.x,a.y-b.y,a.z-b.z);
}

//**************************************************************************

/** a simple scale operator */
template <class T>
inline SomaVector<T> operator * (const SomaVector<T>& a, const T b){
  return SomaVector<T>(a.x*b,a.y*b,a.z*b);
}

//**************************************************************************

/** a simple scale operator */
template <class T>
inline SomaVector<T> operator * (const T b, const SomaVector<T>& a){
  return SomaVector<T>(a.x*b,a.y*b,a.z*b);
}

//**************************************************************************

/** a simple scale operator */
template <class T>
inline SomaVector<T> operator / (const SomaVector<T>& a, const T b){
  return SomaVector<T>(a.x/b,a.y/b,a.z/b);
}

//**************************************************************************

/** a simple scale operator */
template <class T>
inline SomaVector<T> operator / (const T b, const SomaVector<T>& a){
  return SomaVector<T>(b/a.x,b/a.y,b/a.z);
}

//**************************************************************************

/** cross product */
template <class T>
inline SomaVector<T> cross(const SomaVector<T>& a,const SomaVector<T>& b){
  return SomaVector<T>((a.y*b.z) - (a.z*b.y),
                       (a.z*b.x) - (a.x*b.z),
                       (a.x*b.y) - (a.y*b.x));
}

//**************************************************************************

/** dot product */
template <class T>
inline T dot(const SomaVector<T>& a,const SomaVector<T>& b){
  return a.x*b.x + a.y*b.y + a.z*b.z;
}

//**************************************************************************

#endif

